import uuid
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, root_validator


class DBSchema(BaseModel):
    class Config:
        orm_mode = True


class LocalTrainMasterImageBase(BaseModel):
    registry: Optional[str] = None
    group: Optional[str] = None
    artifact: Optional[str] = None
    tag: Optional[str] = "latest"
    image_id: Optional[str] = None

    @root_validator
    def image_specified(cls, values):
        image_id = values.get("image_id")
        if not image_id:
            registry, group, artifact = (
                values.get("registry"),
                values.get("group"),
                values.get("artifact"),
            )
            if not registry or not group or not artifact:
                raise ValueError(
                    "Image ID or registry, group and artifact must be specified."
                )
        return values


class LocalTrainMasterImageCreate(LocalTrainMasterImageBase):
    pass


class LocalTrainMasterImageUpdate(LocalTrainMasterImageBase):
    pass


class LocalTrainMasterImage(LocalTrainMasterImageBase):
    id: uuid.UUID
    created_at: datetime
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True


class LocalTrainConfigurationStep(str, Enum):
    initialized = "initialized"
    image_configured = "image_configured"
    files_uploaded = "files_uploaded"
    dataset_selected = "dataset_selected"
    train_configured = "train_configured"
    finished = "finished"


class LocalTrainStateBase(BaseModel):
    last_execution: Optional[datetime] = None
    num_executions: int = 0
    status: str = "inactive"
    configuration_state: LocalTrainConfigurationStep = None

    class Config:
        use_enum_values = True


class LocalTrainStateCreate(LocalTrainStateBase):
    pass


class LocalTrainStateUpdate(LocalTrainStateBase):
    pass


class LocalTrainState(LocalTrainStateBase):
    id: int

    class Config:
        orm_mode = True


class LocalTrainExecution(BaseModel):
    id: uuid.UUID
    train_id: uuid.UUID
    airflow_dag_run: Optional[str] = None
    config_id: Optional[int] = None
    dataset_id: Optional[uuid.UUID] = None
    start: datetime
    finish: Optional[datetime] = None

    class Config:
        orm_mode = True


class LocalTrainBase(BaseModel):
    name: Optional[str] = None
    master_image_id: Optional[Any] = None
    entrypoint: Optional[str] = None
    files: Optional[List[str]] = None
    custom_image: Optional[str] = None
    fhir_query: Optional[Union[str, dict]] = None
    command: Optional[str] = None
    command_args: Optional[str] = None


class LocalTrainCreate(LocalTrainBase):
    pass


class LocalTrainUpdate(LocalTrainBase):
    pass


class LocalTrain(LocalTrainBase, DBSchema):
    id: uuid.UUID
    created_at: datetime
    updated_at: Optional[datetime] = None
    status: Optional[str] = None
    state: Optional[LocalTrainState] = None


class LocalTrainRunConfig(BaseModel):
    dataset_id: Optional[str] = None
    config_id: Optional[int] = None
    config: Optional[Dict] = None



import tarfile
import time
from io import BufferedReader, BytesIO, TextIOWrapper
from typing import Dict, List, Union
from zipfile import ZipFile

import pendulum
import starlette
from fastapi import File, UploadFile
from loguru import logger
from minio import Minio
from minio.error import S3Error
from pydantic import SecretStr

from station.app.schemas.datasets import MinioFile
from station.app.schemas.station_status import HealthStatus
from station.common.constants import DataDirectories


class MinioClient:
    def __init__(
        self, minio_server: str = None, access_key: str = None, secret_key: str = None
    ):
        """
        Initialize a minio client either with the values passed to the constructor or based on environment variables

        :param minio_server: endpoint of the minio server
        :param access_key: minio access key or username
        :param secret_key: minio password
        """
        # Initialize class fields based on constructor values or environment variables

        # if settings.config.minio.port:
        #     minio_url = f"{settings.config.minio.host}:{settings.config.minio.port}"
        # else:
        #     minio_url = settings.config.minio.host
        # minio_user = settings.config.minio.access_key
        # minio_pass = settings.config.minio.secret_key

        self.minio_server = minio_server
        self.access_key = access_key
        self.secret_key = secret_key

        if isinstance(self.secret_key, SecretStr):
            self.secret_key = self.secret_key.get_secret_value()

        assert self.minio_server
        assert self.access_key
        assert self.secret_key

        # Initialize minio client
        self.client = Minio(
            self.minio_server,
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure=False,
        )

    async def store_model_file(self, id: str, model_file: Union[File, UploadFile]):
        model_data = await model_file.read()
        file = BytesIO(model_data)
        res = self.client.put_object(
            DataDirectories.MODELS.value,
            object_name=id,
            data=file,
            length=len(file.getbuffer()),
        )
        return res

    def get_model_file(self, model_id: str) -> bytes:
        try:
            response = self.client.get_object(
                bucket_name=DataDirectories.MODELS.value, object_name=model_id
            )
            data = response.read()
        finally:
            response.close()
            response.release_conn()

        return data

    async def store_files(self, bucket: str, name: str, file: Union[File, UploadFile]):
        """
        store files into minio
        """
        print(name)
        print(type(file))
        if isinstance(file, BufferedReader):
            model_data = file.read()
        elif isinstance(file, str):
            model_data = file.encode("utf-8")
        elif isinstance(file, TextIOWrapper):
            model_data = file.read().encode("utf-8")
        elif isinstance(file, BytesIO):
            res = self.client.put_object(
                bucket, object_name=name, data=file, length=len(file.getbuffer())
            )
            return res
        elif isinstance(file, starlette.datastructures.UploadFile):
            model_data = await file.read()
        else:
            raise TypeError(f"files with type {type(file)} are not supported")

        file = BytesIO(model_data)
        res = self.client.put_object(
            bucket, object_name=name, data=file, length=len(file.getbuffer())
        )
        return res

    async def save_dataset_files(
        self, dataset_id: str, files: List[Union[File, UploadFile]]
    ):
        """
        store files into minio
        """
        resp = []

        for file in files:
            data = await file.read()

            data_file = BytesIO(data)
            res = self.client.put_object(
                bucket_name=DataDirectories.DATASETS.value,
                object_name=f"{dataset_id}/{file.filename}",
                data=data_file,
                length=len(data),
                content_type="text/plain",
            )
            resp.append(res)
            data_file.seek(0)

        return resp

    async def save_local_train_files(
        self, train_id: str, files: List[Union[File, UploadFile]]
    ) -> List[MinioFile]:
        """
        store files of local train in minio
        """
        resp = []

        for file in files:
            data = await file.read()
            data_file = BytesIO(data)
            res = self.client.put_object(
                bucket_name=DataDirectories.LOCAL_TRAINS.value,
                object_name=f"{train_id}/{file.filename}",
                data=data_file,
                length=len(data),
            )

            resp.append(
                MinioFile(
                    file_name=file.filename,
                    file_size=len(data),
                    full_path=f"{res.bucket_name}/{res.object_name}",
                    updated_at=pendulum.now(),
                )
            )
        return resp

    def get_local_train_archive(self, train_id: str):
        print(DataDirectories.LOCAL_TRAINS.value)
        items = self.get_minio_dir_items(
            bucket=DataDirectories.LOCAL_TRAINS.value, directory=train_id
        )

        archive = self.make_download_archive(
            DataDirectories.LOCAL_TRAINS.value, items=items
        )

        return archive

    def get_file(self, bucket: str, name: str) -> bytes:
        response = self.client.get_object(bucket_name=bucket, object_name=name)
        data = response.read()
        response.close()
        response.release_conn()

        return data

    def delete_file(self, bucket: str, name: str):
        self.client.remove_object(bucket_name=bucket, object_name=name)

    def delete_folder(self, bucket: str, directory: str):
        delete_objects = self.client.list_objects(
            bucket_name=bucket, prefix=directory, recursive=True
        )
        for obj in delete_objects:
            self.client.remove_object(bucket_name=bucket, object_name=obj.object_name)

    def get_file_names(self, bucket: str, prefix: str = "") -> List[str]:
        response = self.client.list_objects(bucket, prefix=prefix)
        data = list(response)
        return data

    def add_bucket(self, bucket_name: str):
        found = self.client.bucket_exists(bucket_name)
        if not found:
            self.client.make_bucket(bucket_name)

    def list_data_sets(self):
        data_sets = self.client.list_objects("datasets")
        return [ds.object_name for ds in list(data_sets)]

    def list_buckets(self):
        buckets = self.client.list_buckets()
        return buckets

    def list_elements_in_bucket(self, bucket):
        elements = self.client.list_objects(bucket)
        return [ds.object_name for ds in list(elements)]

    def load_data_set(self):
        pass

    def get_minio_dir_items(
        self, bucket: Union[str, DataDirectories], directory: str
    ) -> List[MinioFile]:
        """
        Get all objects in the data set specified by data_set_id and return them as a generator
        Args:
            bucket:
            directory:

        Returns:

        """
        if isinstance(bucket, DataDirectories):
            bucket = bucket.value
        items = self.client.list_objects(bucket, prefix=directory, recursive=True)
        try:
            dir_files = []
            for item in items:
                print(item)
                dir_files.append(
                    MinioFile(
                        file_name=item.object_name.split("/")[-1],
                        full_path=item.object_name,
                        size=item.size,
                        updated_at=item.last_modified,
                    )
                )
            return dir_files
        except Exception as e:
            print(e)
            return []

    def make_dataset_archive(
        self, data_set_id: str, items: List[MinioFile] = None, archive_type: str = "tar"
    ) -> BytesIO:
        """
        Create an archive of the data set specified by data_set_id and return it as a BytesIO object
        Args:
            data_set_id: id of the dataset
            items: list of items to include in the archive
            archive_type: tar or zip

        Returns:

        """
        if items is None:
            items = self.get_minio_dir_items("datasets", data_set_id)

        archive = self.make_download_archive(
            "datasets", items, archive_type=archive_type
        )

        return archive

    def make_download_archive(
        self, bucket: str, items: List[MinioFile], archive_type: str = "tar"
    ) -> BytesIO:
        archive = BytesIO()

        if archive_type == "tar":
            with tarfile.TarFile(fileobj=archive, mode="w") as tar:
                for file in items:
                    data = self.get_file(bucket, file.full_path)
                    info = tarfile.TarInfo(name=file.full_path)
                    info.size = len(data)
                    info.mtime = time.time()
                    tar.addfile(info, BytesIO(data))

            archive.seek(0)
            return archive
        elif archive_type == "zip":
            with ZipFile(archive, "w") as zip:
                for item in items:
                    data = self.get_file(bucket, item.full_path)
                    zip.writestr(item.full_path, data)
            archive.seek(0)
            return archive

        raise ValueError(f"Unknown archive type {archive_type}")

    def get_classes_by_folders(self, data_set_id: str) -> List[str]:
        """
        Gets the subdirectories of a dataset directory in minio. The folder names correspond the classes defined for
        the dataset

        :param data_set_id: identifier of the data set in the datasets bucket
        :return: List of directory (class) names found in the specified directory
        """

        folders = self.client.list_objects(
            "datasets", prefix=data_set_id, recursive=False
        )
        classes = []
        for folder in folders:
            classes.append(folder.object_name.split("/")[-2])
        return classes

    def get_class_distributions(
        self, data_set_id: str, classes: List[str]
    ) -> List[Dict[str, Union[int, str]]]:
        class_distribution = []

        for cls in classes:
            prefix = data_set_id + cls + "/"
            class_items = len(list(self.client.list_objects("datasets", prefix=prefix)))
            class_object = {"class_name": cls, "n_items": class_items}
            class_distribution.append(class_object)

        return class_distribution

    def health_check(self) -> HealthStatus:
        """
        Get health of minio
        """
        try:
            self.client.list_buckets()

            return HealthStatus.healthy
        except Exception as e:
            logger.error(f"Error while checking minio health: {e}")
            return HealthStatus.error

    def setup_buckets(self):
        for d in DataDirectories:
            try:
                self.client.make_bucket(d.value)
            except S3Error as se:
                if se.code == "BucketAlreadyOwnedByYou":
                    pass
                else:
                    raise se
            except Exception as e:
                raise e



from typing import Any

import requests

from station.common.clients.base import BaseClient
from station.common.clients.central.schemas import RegistryCredentials


class CentralApiClient(BaseClient):
    def __init__(self, base_url: str, robot_id: str, robot_secret: str):
        super().__init__(
            base_url=base_url,
            robot_id=robot_id,
            robot_secret=robot_secret,
            auth_url=f"{base_url}/auth",
        )
        self.base_url = base_url

        self.api_url = base_url + "/api"

    def get_trains(self, station_id: Any) -> dict:
        url = self.api_url + "/train-stations?"
        filters = f"filter[station_id]={station_id}&include=train"
        safe_filters = self._make_url_safe(filters)
        url = url + safe_filters
        response = requests.get(url, headers=self.headers)
        response.raise_for_status()
        return response.json()

    def get_registry_credentials(self, station_id: Any) -> RegistryCredentials:
        # get registry and external name
        url = self.api_url + "/stations?"
        filters = (
            "fields=+secure_id,+registry_project_account_name,"
            "+registry_project_account_token&"
            f"filter[id]={station_id}"
        )
        safe_filters = self._make_url_safe(filters)
        url = url + safe_filters
        r = requests.get(url, headers=self.headers)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            print(r.content)
            raise e
        # create credentials object
        station_registry_data = r.json()["data"][0]
        # extract station associated data from initial response
        project = station_registry_data["external_name"]
        registry_id = station_registry_data["registry_id"]
        registry_project_id = station_registry_data["registry_project_id"]

        address = self._get_registry_url(registry_id)
        user, password = self._get_registry_project_credentials(registry_project_id)

        credentials = RegistryCredentials(
            address=address,
            user=user,
            password=password,
            project=project,
        )
        return credentials

    def _get_registry_url(self, registry_id: Any) -> str:
        url = self.api_url + f"/registries/{registry_id}"
        r = requests.get(url, headers=self.headers)
        r.raise_for_status()
        return r.json()["host"]

    def _get_registry_project_credentials(self, registry_project_id: str) -> tuple:
        url = self.api_url + f"/registry-projects/{registry_project_id}?"
        filters = "fields=+account_id,+account_name,+account_secret"
        safe_filters = self._make_url_safe(filters)
        url = url + safe_filters
        r = requests.get(url, headers=self.headers)
        try:
            r.raise_for_status()
        except Exception as e:
            print(r.content)
            raise e
        return r.json()["account_name"], r.json()["account_secret"]

    def update_public_key(self, station_id: Any, public_key: str) -> dict:
        url = self.api_url + f"/stations/{station_id}"
        payload = {"public_key": public_key}
        r = requests.post(url, headers=self.headers, json=payload)
        r.raise_for_status()
        return r.json()



from typing import Any, Generic, List, Type, TypeVar, Union

import requests
import requests.auth
from pydantic import BaseModel
from requests import HTTPError

from station.common.clients.base import BaseClient

ModelType = TypeVar("ModelType", bound=BaseModel)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class ResourceClient(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(
        self, base_url, resource_name: str, model: Type[ModelType], client: BaseClient
    ):
        """
        Base class for a resource client for the station api implement
        Args:
            base_url:
            resource_name:
            model:
            client: Base client to handle authentication
        """
        self.base_url = base_url
        self.resource_name = resource_name
        self.model = model
        self._client = client

    def create(self, data: Union[CreateSchemaType, dict]) -> ModelType:
        if isinstance(data, dict):
            data = self.model(**data)
        response = requests.post(
            f"{self.base_url}/{self.resource_name}",
            json=data.dict(),
            headers=self._client.headers,
        )
        try:
            response.raise_for_status()
        except HTTPError as e:
            print(response.text)
            raise e
        return self.model(**response.json())

    def get(self, resource_id) -> ModelType:
        response = requests.get(
            f"{self.base_url}/{self.resource_name}/{resource_id}",
            headers=self._client.headers,
        )
        response.raise_for_status()
        return self.model(**response.json())

    def get_multi(self, skip: int = 0, limit: int = 100) -> List[ModelType]:
        response = requests.get(
            f"{self.base_url}/{self.resource_name}",
            params={"skip": skip, "limit": limit},
            headers=self._client.headers,
        )
        response.raise_for_status()
        return [self.model(**item) for item in response.json()]

    def update(self, resource_id: Any, data: UpdateSchemaType) -> ModelType:
        response = requests.put(
            f"{self.base_url}/{self.resource_name}/{resource_id}",
            json=data,
            headers=self._client.headers,
        )
        response.raise_for_status()
        return self.model(**response.json())

    def delete(self, resource_id) -> ModelType:
        response = requests.delete(
            f"{self.base_url}/{self.resource_name}/{resource_id}",
            headers=self._client.headers,
        )
        response.raise_for_status()
        return self.model(**response.json())



import os

import click
from rich.console import Console

from station.common.constants import Icons
from station.ctl.config import find_config, fix_config, load_config
from station.ctl.config.validate import validate_config
from station.ctl.util import get_template_env


@click.command(help="Validate and/or fix a station configuration file")
@click.option("-f", "--file", help="Path to the configuration file to validate/fix")
@click.option(
    "--dry-run",
    is_flag=True,
    help="Do not write the fixed config to disk. But print it instead.",
)
@click.pass_context
def settings(ctx, file, dry_run):
    """Validate and/or fix the configuration file"""

    if not file:
        click.echo(
            "No configuration file specified. Looking for a config file in the current directory... ",
            nl=False,
        )
        station_config, file = find_config(os.getcwd())
        click.echo(f"{Icons.CHECKMARK.value}")
    else:
        station_config = load_config(file)

    click.echo("Validating configuration file...")

    results = validate_config(station_config)

    if results is not None:
        table, results = results
        console = Console()
        console.print(table)
        click.confirm("Fix issues now?", abort=True)
        fixed_config = fix_config(ctx.obj, station_config, results)
        render_config(fixed_config, file)
        if not dry_run:
            click.echo(f"Fixed configuration file written to: {file}")

    else:
        click.echo("Configuration file is valid.")


def render_config(config: dict, path: str, dry_run: bool = False) -> str | None:
    env = get_template_env()
    template = env.get_template("station_config.yml.tmpl")
    # write out the correct path to key file on host when rendering the template from docker container

    # todo check and improve this
    if config.get("host_path"):
        key_name = config["station_config"]["central"]["private_key"].split("/")[-1]
        key_path = os.path.join(config["host_path"], key_name)
        config["station_config"]["central"]["private_key"] = key_path

    # todo fix this hack
    # extract https certs from config
    certs = config["station_config"]["https"].pop("certificate")

    out_config = template.render(certificate=certs, **config["station_config"])

    # print the rendered config to stdout and return it if dry_run is True
    if dry_run:
        click.echo(out_config)
        return out_config
    else:
        with open(path, "w") as f:
            f.write(out_config)



from datetime import datetime
from typing import Any, Dict, Union

from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from station.app.config import clients
from station.app.crud.base import (
    CreateSchemaType,
    CRUDBase,
    ModelType,
    UpdateSchemaType,
)
from station.app.models.local_trains import (
    LocalTrain,
    LocalTrainExecution,
    LocalTrainState,
)
from station.app.schemas.local_trains import (
    LocalTrainConfigurationStep,
    LocalTrainCreate,
    LocalTrainUpdate,
)
from station.common.constants import DataDirectories
from station.trains.local.update import update_configuration_status


class CRUDLocalTrain(CRUDBase[LocalTrain, LocalTrainCreate, LocalTrainUpdate]):
    def create(self, db: Session, *, obj_in: CreateSchemaType) -> ModelType:
        obj_in_data = jsonable_encoder(obj_in, exclude_none=True)
        db_obj = self.model(**obj_in_data)
        db.add(db_obj)
        db.commit()
        train = self.create_initial_state(db, db_obj)
        state = train.state
        state.configuration_state = LocalTrainConfigurationStep.initialized.value
        db.commit()

        return train

    def create_run(
        self,
        db: Session,
        *,
        train_id: str,
        dag_run: str,
        config_id: int = None,
        dataset_id: str = None
    ) -> LocalTrainExecution:
        """
        create a database entry for a local train execution

        @param db: eference to the postgres database
        @param obj_in: LocalTrainRun json as defind in the schemas
        @return: local run object
        """
        run = LocalTrainExecution(
            train_id=train_id,
            airflow_dag_run=dag_run,
            config_id=config_id,
            dataset_id=dataset_id,
        )
        db.add(run)
        db.commit()
        db.refresh(run)
        return run

    async def remove_train(self, db: Session, train_id: str) -> ModelType:
        """

        @param db:
        @param train_id:
        @return:
        """
        # TODO remove query results when exist
        # remove files stored in minio
        clients.minio.delete_folder(
            bucket=str(DataDirectories.LOCAL_TRAINS.value), directory=train_id
        )

        # remove sql database entries for LocalTrainExecution
        obj = (
            db.query(LocalTrainExecution)
            .filter(LocalTrainExecution.train_id == train_id)
            .all()
        )
        for run in obj:
            db.delete(run)
        db.commit()
        # remove sql database entry for LocalTrain
        db_train = db.query(LocalTrain).filter(LocalTrain.train_id == train_id).first()
        db.delete(db_train)
        db.commit()
        return db_train

    def create_initial_state(self, db: Session, db_obj: LocalTrain):
        state = LocalTrainState(train_id=db_obj.id)
        db.add(state)
        db.commit()
        db.refresh(db_obj)
        return db_obj

    def update(
        self,
        db: Session,
        *,
        db_obj: ModelType,
        obj_in: Union[UpdateSchemaType, Dict[str, Any]]
    ) -> ModelType:
        update_train = super().update(db, db_obj=db_obj, obj_in=obj_in)
        update_train.updated_at = datetime.now()
        db.commit()
        state = update_train.state
        files = clients.minio.get_minio_dir_items(
            DataDirectories.LOCAL_TRAINS.value, db_obj.id
        )
        configuration_state = update_configuration_status(update_train, files)
        state.configuration_state = configuration_state
        db.commit()
        return update_train


local_train = CRUDLocalTrain(LocalTrain)



from .crud_datasets import datasets
from .crud_docker_trains import docker_trains
from .crud_fhir_servers import fhir_servers
from .crud_local_train import local_train
from .crud_notifications import notifications

# from .crud_discovery import discoveries
# TODO cleanup namespace inconsistencies



from typing import List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.crud.local_train_master_image import local_train_master_image
from station.app.schemas import local_trains

router = APIRouter()


@router.post("", response_model=local_trains.LocalTrainMasterImage)
def add_master_image(
    add_master_image_msg: local_trains.LocalTrainMasterImageCreate,
    db: Session = Depends(dependencies.get_db),
):
    # check if id already exists
    if local_train_master_image.get_by_image_id(db, add_master_image_msg.image_id):
        raise HTTPException(
            status_code=400,
            detail=f"Image with the given id: ({add_master_image_msg.image_id}) already exists",
        )
    image = local_train_master_image.create(db, obj_in=add_master_image_msg)
    return image


@router.get("/{image_id}", response_model=local_trains.LocalTrainMasterImage)
def get_master_image(image_id: str, db: Session = Depends(dependencies.get_db)):
    image = local_train_master_image.get(db, image_id)
    if not image:
        raise HTTPException(status_code=404, detail="Image not found")
    return image


@router.get("", response_model=List[local_trains.LocalTrainMasterImage])
def list_master_images(
    db: Session = Depends(dependencies.get_db),
    skip: int = 0,
    limit: int = 100,
    sync: bool = False,
):
    if sync:
        local_train_master_image.sync_with_harbor(db)
    images = local_train_master_image.get_multi(db, skip=skip, limit=limit)
    return images


@router.put("/{image_id}", response_model=local_trains.LocalTrainMasterImage)
def update_master_image(
    image_id: str,
    update_msg: local_trains.LocalTrainMasterImageUpdate,
    db: Session = Depends(dependencies.get_db),
):
    db_image = local_train_master_image.get(db, image_id)
    if not db_image:
        raise HTTPException(status_code=404, detail="Image not found")
    image = local_train_master_image.update(db, db_obj=db_image, obj_in=update_msg)
    return image


@router.delete("/{image_id}", response_model=local_trains.LocalTrainMasterImage)
def delete_master_image(image_id: str, db: Session = Depends(dependencies.get_db)):
    db_image = local_train_master_image.get(db, image_id)
    if not db_image:
        raise HTTPException(status_code=404, detail="Image not found")
    image = local_train_master_image.remove(db, db_obj=db_image)
    return image



from typing import Any, Callable

from pydantic import BaseModel

from station.common.config.generators import (
    generate_fernet_key,
    generate_private_key,
    password_generator,
)


class GeneratorArg(BaseModel):
    prompt: str
    arg: str
    required: bool = True
    value: Any | None = None


class ConfigItemFix(BaseModel):
    issue: str
    value: Any | None = None
    suggestion: str
    fix: Any | None = None
    generator_function: Callable | None = None
    generator_args: list[GeneratorArg] | None = None

    @classmethod
    def no_fix(
        cls,
        suggestion: str,
        value: str | None = None,
    ) -> "ConfigItemFix":
        return cls(
            value=value,
            issue="",
            fix=None,
            suggestion=suggestion,
        )

    @classmethod
    def admin_password(cls, value: str) -> "ConfigItemFix":
        suggested_password = password_generator()
        return cls(
            issue=f"Invalid admin password [{value}]",
            value=value,
            suggestion=f'Use suggested password: "{suggested_password}"',
            fix=suggested_password,
        )

    @classmethod
    def private_key(cls, value: str) -> "ConfigItemFix":
        """
        Fix that allow for the generation of a new private key
        """

        # Generator arguments to generate a new private key
        args = [
            GeneratorArg(
                prompt="Enter the directory in which to save the private key",
                arg="path",
            ),
            GeneratorArg(
                prompt="Enter the name of the private key file",
                arg="name",
            ),
            GeneratorArg(
                prompt="Optionally Enter a passphrase to encrypt the private key",
                arg="password",
                required=False,
            ),
        ]
        return cls(
            issue="Invalid private key",
            value=value,
            suggestion="Either generate a new private key yourself or enter GENERATE to generate a new one",
            fix="GENERATE",
            generator_function=generate_private_key,
            generator_args=args,
        )

    @classmethod
    def fernet_key(cls, value: str) -> "ConfigItemFix":
        return cls(
            issue="Invalid fernet key",
            value=value,
            suggestion="Enter a valid fernet key or GENERATE to generate a new one",
            fix="GENERATE",
            generator_function=generate_fernet_key,
        )



from sqlalchemy import JSON, Column, Integer

from station.app.db.base_class import Base


class DataSetSummary(Base):
    __tablename__ = "datasets_summary"
    id = Column(Integer, primary_key=True, index=True)
    proposal_id = Column(Integer, default=0)
    count = Column(Integer, default=0)
    data_information = Column(JSON, default={})



import copy
import os
import pathlib
from unittest import mock

import pytest
from pydantic import ValidationError

from station.common.config.station_config import StationConfig
from station.ctl.config.command import render_config

TEST_DIR = pathlib.Path(__file__).parent

CONFIG_DICT = {
    "id": "d7b7bd69-c828-45f3-b0bc-0d5ca10a8cd5",
    "version": "latest",
    "environment": "development",
    "admin_password": "GucVuG6MgVyy58v8Xjg3o4jTnAyNrP1k",
    "data_dir": str(TEST_DIR / "station_data"),
    "central": {
        "api_url": "https://dev.personalhealthtrain.de/api",
        "robot_id": "0b34acb9-9b26-4780-80d1-705772464cf2",
        "robot_secret": "start123",
        "private_key": "./private.pem",
        "private_key_password": "",
    },
    "http": {"port": "80"},
    "https": {
        "port": "443",
        "domain": "station.localhost",
        "certificate": {
            "cert": str(TEST_DIR / "cert.pem"),
            "key": str(TEST_DIR / "key.pem"),
        },
    },
    "traefik": {"dashboard": True, "dashboard_port": 8081},
    "registry": {
        "address": "dev-harbor.personalhealthtrain.de",
        "user": "test-user",
        "password": "test-password",
        "project": "test-project",
    },
    "db": {"host": "127.0.0.1", "admin_user": "admin", "admin_password": "start123"},
    "api": {"fernet_key": "Z_kebTcA7p2VV9xga-ES2wCMjvfaRNzQktjsxo5vPMM="},
    "airflow": {
        "host": "127.0.0.1",
        "admin_user": "admin",
        "admin_password": "start123",
    },
    "auth": {
        "admin_user": "admin",
        "host": "127.0.0.1",
        "port": 3001,
        "robot_id": "5f77fe3f-a48c-46be-ba67-df6b1058ebcb",
        "robot_secret": "v1ziczynshyotuzudc8ymumjtijli9yoo1mygyrv2ucqdm77lae6d5pni6xh5vp4",
    },
    "minio": {
        "host": "127.0.0.1",
        "port": 9000,
        "admin_user": "minio_admin",
        "admin_password": "minio_admin",
    },
    "redis": {
        "host": "127.0.0.1",
        "admin_password": "start123",
    },
}


def test_config_from_dict():
    config = StationConfig(**CONFIG_DICT)
    config.display()


def test_construct():
    modified_config_dict = copy.deepcopy(CONFIG_DICT)
    modified_config_dict["https"]["certificate"]["key"] = "test-id"
    config = StationConfig.construct(**modified_config_dict)

    config.display()

    assert config.https.certificate.key == "test-id"

    with pytest.raises(ValidationError):
        config = StationConfig(modified_config_dict)


def test_from_file(tmp_path):
    # create temp path
    config_file = tmp_path / "test_station_config.yml"

    # render config file and write to path
    render_config(CONFIG_DICT, path=str(config_file))

    # print config file
    print(config_file.read_text())

    config = StationConfig.from_file(str(config_file))
    config.display()


def test_env_vars():
    # mock env vars

    # test top level env var
    with mock.patch.dict(os.environ, {"STATION_ID": "env-test"}):
        print(os.getenv("STATION_ID"))
        config = StationConfig(**CONFIG_DICT)
        assert config.id == "env-test"

    # test nested env var
    with mock.patch.dict(os.environ, {"STATION_MINIO_HOST": "env-test"}):
        config = StationConfig(**CONFIG_DICT)
        assert config.minio.host == "env-test"

    # test nested env var
    with mock.patch.dict(os.environ, {"STATION_MINIO_HOST": "env-test"}):
        config = StationConfig.construct(**CONFIG_DICT)
        assert config.minio.host == "env-test"



from fastapi.testclient import TestClient

from station.app.api.dependencies import get_db
from station.app.main import app

from .test_db import override_get_db

app.dependency_overrides[get_db] = override_get_db

client = TestClient(app)


def test_data_set_create():
    response = client.post(
        "/api/datasets",
        json={"name": "test data set", "data_type": "tabular", "storage_type": "fhir"},
    )

    assert response.status_code == 200, response.text

    data = response.json()
    dataset_id = data["id"]
    assert dataset_id


def test_data_set_get():
    response = client.get(f"/api/datasets/{1}")
    assert response.status_code == 200, response.text

    data = response.json()
    assert data["name"] == "test data set"


def test_get_all_data_sets():
    response = client.get("/api/datasets")
    assert response.status_code == 200, response.text
    data = response.json()

    assert len(data) >= 1



from enum import Enum
from typing import List, Tuple

from pydantic import BaseModel, ValidationError
from rich.style import Style
from rich.table import Table

from station.common.config.fix import ConfigItemFix
from station.common.config.station_config import StationConfig
from station.ctl.config.fix import get_fixes_from_errors


class Colors(Enum):
    YELLOW = Style(color="yellow")
    RED = Style(color="red")


class ValidationResult(BaseModel):
    loc: tuple
    fix: ConfigItemFix


def validate_config(
    config: dict, host_path: str = None
) -> Tuple[Table, List[ValidationResult]] | None:
    try:
        StationConfig(**config)
        return None
    except ValidationError as e:
        errors = e.errors()
        fixes = get_fixes_from_errors(config, errors)

        results = []

        for e, fix in zip(errors, fixes):
            fix.issue = e["msg"]
            results.append(
                ValidationResult(
                    loc=e["loc"],
                    fix=fix,
                )
            )

        table = _generate_results_table(results)

        return table, results
    except Exception as e:
        print("Unknown exception while validating config")
        print(e)
        raise e


def _generate_results_table(results: List[ValidationResult]) -> Table:
    table = Table(
        title="Staion config validation results", show_lines=True, header_style="bold"
    )
    table.add_column("Loc", justify="center")
    table.add_column("Issue", justify="center")
    table.add_column("Hint", justify="center")
    table.add_column("Fix", justify="center")

    for result in results:
        table.add_row(
            ".".join(result.loc),
            result.fix.issue,
            result.fix.suggestion,
            result.fix.fix,
        )

    return table



from dotenv import find_dotenv, load_dotenv
from fastapi import Depends, FastAPI
from fastapi.middleware.cors import CORSMiddleware

from station.app.api.api_v1.api import api_router
from station.app.auth import authorized_user
from station.app.logger import init_logging

load_dotenv(find_dotenv())

app = FastAPI(
    title="PHT Station",
    docs_url="/api/docs",
    redoc_url="/api/redoc",
    openapi_url="/api/v1/openapi.json",
)

# Setup logging
init_logging()

# TODO remove full wildcard for production
origins = [
    "http://localhost:8080",
    "http://localhost:8080/",
    "http://localhost:8081",
    # "http://localhost:3000",
    # "http://localhost",
    "*",
]


app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(
    api_router,
    prefix="/api",
    dependencies=[Depends(authorized_user)],
)



from typing import Any, Dict, Union

from sqlalchemy.orm import Session

from station.app.models.fhir_server import FHIRServer
from station.app.schemas.fhir import FHIRServerCreate, FHIRServerUpdate

from .base import CreateSchemaType, CRUDBase, ModelType, UpdateSchemaType


class CRUDFHIRServers(CRUDBase[FHIRServer, FHIRServerCreate, FHIRServerUpdate]):
    def create(self, db: Session, *, obj_in: CreateSchemaType) -> ModelType:
        # Encrypt sensitive values
        encrypted_in = self.encrypt_sensitive_values(obj_in)
        # Add encrypted values to the db
        return super().create(db, obj_in=encrypted_in)

    def update(
        self,
        db: Session,
        *,
        db_obj: ModelType,
        obj_in: Union[UpdateSchemaType, Dict[str, Any]]
    ) -> ModelType:
        obj_in = self.encrypt_sensitive_values(obj_in)
        return super().update(db, db_obj=db_obj, obj_in=obj_in)

    def encrypt_sensitive_values(
        self, obj_in: Union[FHIRServerCreate, FHIRServerUpdate]
    ) -> FHIRServerCreate:
        """
        Encrypt sensitive values in the FHIRServerCreate/FHIRServerUpdate object
        Args:
            obj_in:
        Returns:
            the same object with encrypted sensitive values
        """

        from station.app.settings import settings

        fernet = settings.get_fernet()

        if obj_in.client_secret:
            obj_in.client_secret = fernet.encrypt(
                obj_in.client_secret.encode()
            ).decode()
        if obj_in.password:
            obj_in.password = fernet.encrypt(obj_in.password.encode()).decode()
        if obj_in.token:
            obj_in.token = fernet.encrypt(obj_in.token.encode()).decode()

        return obj_in


fhir_servers = CRUDFHIRServers(FHIRServer)



import os
from typing import Any, Dict

import requests
from sqlalchemy.orm import Session


class ConductorRESTClient:
    def __init__(self, conductor_url: str = None, station_id: int = None):
        self.conductor_url = (
            conductor_url if conductor_url else os.getenv("CONDUCTOR_URL")
        )
        self.station_id = station_id if station_id else os.getenv("STATION_ID")

        assert self.conductor_url
        assert self.station_id

    def get_model_for_train(self, train_id: Any, db: Session):
        """
        Obtain the model definition for the train given by the id from the conductor

        :param train_id:
        :param db:
        :return:
        """

        url = self.conductor_url + f"/api/trains/{train_id}/model"
        r = requests.get(url=url)
        return r.json()

    def upload_model_parameters(self, train_id: Any):
        pass

    def get_aggregated_model(self, train_id: Any):
        pass

    def get_available_trains(self):
        url = self.conductor_url + f"/api/stations/{self.station_id}/trains"
        r = requests.get(url)
        r.raise_for_status()
        return r.json()

    def post_discovery_results(self, train_id: Any, discovery_results: Dict):
        url = self.conductor_url + f"/api/trains/{train_id}/discovery"
        r = requests.post(url, json=discovery_results)
        print(r.request.body)
        r.raise_for_status()
        return r.json()



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class DBSchema(BaseModel):
    class Config:
        orm_mode = True


class NotificationBase(BaseModel):
    topic: Optional[str] = None
    message: Optional[str] = None
    target_user: Optional[str] = "all"
    title: Optional[str] = None
    is_read: Optional[bool] = None
    type: Optional[str] = None


class NotificationCreate(NotificationBase):
    pass


class NotificationUpdate(NotificationBase):
    pass


class Notification(NotificationBase, DBSchema):
    id: int
    message: str
    topic: Optional[str] = "trains"
    target_user: Optional[str] = "all"
    is_read: bool
    created_at: datetime



import os

import requests
from loguru import logger
from requests.auth import HTTPBasicAuth

from station.app.schemas.station_status import HealthStatus


class AirflowClient:
    def __init__(
        self,
        airflow_api_url: str = None,
        airflow_user: str = None,
        airflow_password: str = None,
    ):
        self.airflow_url = (
            airflow_api_url
            if airflow_api_url
            else os.getenv("AIRFLOW_API_URL", "localhost:8080/api/v1")
        )
        self.airflow_user = (
            airflow_user if airflow_user else os.getenv("AIRFLOW_USER", "admin")
        )
        self.airflow_pw = (
            airflow_password if airflow_password else os.getenv("AIRFLOW_PW", "admin")
        )
        self.auth = HTTPBasicAuth(self.airflow_user, self.airflow_pw)

    def trigger_dag(self, dag_id: str, config: dict = None) -> str:
        """
        Execute a dag with the given configuration

        :param dag_id: string identifier of the dag i.e. 'run_train'
        :param config: dictionary containing configuration values passed to the dag
        :return:
        """

        # create default empty config
        config_msg = {
            "conf": {},
        }
        if config:
            config_msg["conf"] = config

        url = self.airflow_url + f"dags/{dag_id}/dagRuns"
        r = requests.post(url=url, auth=self.auth, json=config_msg)
        try:
            r.raise_for_status()

        except Exception as e:
            logger.error(f"Error triggering dag: \n{e}")
            logger.error(f"Error message: \n{r.text}")
            raise e
        return r.json()["dag_run_id"]

    def get_dag_run(self, dag_id: str):
        pass

    def get_all_dag_runs(self, dag_id: str):
        url = self.airflow_url + f"dags/{dag_id}/dagRuns"
        r = requests.get(url=url, auth=self.auth)
        r.raise_for_status()
        return r.json()

    def get_dags(self):
        url = self.airflow_url + "dags"
        r = requests.get(url=url, auth=self.auth)
        r.raise_for_status()

        return r.json()

    # TODO create arguments for individual connection options
    def create_connection(self, connection_dict: dict):
        url = self.airflow_url + "connections"
        r = requests.post(url=url, json=connection_dict)
        r.raise_for_status()

    def health_check(self) -> HealthStatus:
        """

        @return: dict: Airflow Status
        """
        url = self.airflow_url + "/health"
        r = requests.get(url=url)
        try:
            r.raise_for_status()
            return HealthStatus.healthy
        except Exception as e:
            logger.error(f"Error checking airflow health: \n{e}")
            return HealthStatus.error

    def get_run_information(self, dag_id: str, run_id: str) -> dict:
        """
        requests the information about a dag run state from airflow.
        @param dag_id: ID of the dag (e.g. "run_train" or "run_local")
        @param run_id: Airflow ID of the run (has the form of e.g. " "manual__2021-11-09T14:12:24.622670+00:00")
        @return: dict: information about the run
        """
        url = self.airflow_url + f"dags/{dag_id}/dagRuns/{run_id}/taskInstances"
        task_list = requests.get(url=url, auth=self.auth)
        task_list.raise_for_status()
        task_list = task_list.json()
        url = self.airflow_url + f"dags/{dag_id}/dagRuns/{run_id}"
        information = requests.get(url=url, auth=self.auth)
        information.raise_for_status()
        information = information.json()
        information["tasklist"] = task_list
        return information

    def get_task_log(
        self, dag_id: str, run_id: str, task_id: str, task_try_number: int = None
    ) -> str:
        """
        get the log of a task for a specific run
        @param dag_id: dag under which the task is run
        @param run_id: id for specific run
        @param task_id: task for which the logs are requested
        @param task_try_number: specific try number of the task, default is the last try
        @return: logs of the task run
        """

        task_number_list = []
        task_known = False
        for task in self.get_run_information(dag_id, run_id)["tasklist"][
            "task_instances"
        ]:
            if task["task_id"] == task_id:
                task_known = True
                task_number_list.append(task["try_number"])
        if task_known:
            last_task_try_number = max(task_number_list)
            if task_try_number and task_try_number <= last_task_try_number:
                try_number = task_try_number
            else:
                try_number = last_task_try_number
            url = (
                self.airflow_url
                + f"dags/{dag_id}/dagRuns/{run_id}/taskInstances/{task_id}/logs/{try_number}"
            )
            log = requests.get(url=url, auth=self.auth)
            log.raise_for_status()
            return log.content.decode("utf-8")
        else:
            return ""


airflow_client = AirflowClient()



from datetime import datetime

from sqlalchemy import JSON, Boolean, Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import relationship

from station.app.db.base_class import Base


class DockerTrainState(Base):
    __tablename__ = "docker_train_states"
    id = Column(Integer, primary_key=True, index=True)
    train_id = Column(Integer, ForeignKey("docker_trains.id"))
    # train = relationship("DockerTrain", backref=backref("state", uselist=False))
    last_execution = Column(DateTime, nullable=True)
    num_executions = Column(Integer, default=0)
    status = Column(String, default="inactive")
    central_status = Column(String, nullable=True)


class DockerTrainExecution(Base):
    __tablename__ = "docker_train_executions"
    id = Column(Integer, primary_key=True, index=True)
    train_id = Column(Integer, ForeignKey("docker_trains.id"))
    # train_state_id = Column(Integer, ForeignKey('docker_train_states.id'), nullable=True)
    start = Column(DateTime, default=datetime.now())
    end = Column(DateTime, nullable=True)
    airflow_dag_run = Column(String, nullable=True)
    config = Column(Integer, ForeignKey("docker_train_configs.id"), nullable=True)
    dataset_id = Column(UUID(as_uuid=True), ForeignKey("datasets.id"), nullable=True)


class DockerTrainConfig(Base):
    __tablename__ = "docker_train_configs"
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, unique=True)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, nullable=True)
    airflow_config = Column(JSON, nullable=True)
    trains = relationship("DockerTrain")
    cpu_requirements = Column(JSON, nullable=True)
    gpu_requirements = Column(JSON, nullable=True)
    auto_execute = Column(Boolean, default=False)
    dataset_id = Column(UUID(as_uuid=True), ForeignKey("datasets.id"), nullable=True)


class DockerTrain(Base):
    __tablename__ = "docker_trains"
    id = Column(Integer, primary_key=True, index=True)
    train_id = Column(String, unique=True, index=True)
    name = Column(String, nullable=True)
    proposal = Column(String, nullable=True)
    is_active = Column(Boolean, default=False)
    image_name = Column(String, nullable=True)
    type = Column(String, nullable=True)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, nullable=True)
    config_id = Column(Integer, ForeignKey("docker_train_configs.id"), nullable=True)
    config = relationship("DockerTrainConfig", back_populates="trains")
    state = relationship("DockerTrainState", uselist=False, cascade="all, delete")
    executions = relationship("DockerTrainExecution", cascade="all, delete")
    num_participants = Column(Integer, nullable=True)



from typing import List

from fastapi import APIRouter, Depends
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.crud.crud_notifications import notifications
from station.app.schemas.notifications import (
    Notification,
    NotificationCreate,
    NotificationUpdate,
)

router = APIRouter()


@router.post("", response_model=Notification, status_code=201)
def add_notification(
    create_msg: NotificationCreate, db: Session = Depends(dependencies.get_db)
):
    db_notification = notifications.create(db=db, obj_in=create_msg)
    return db_notification


@router.put("/{notification_id}", response_model=Notification)
def update_notification(
    notification_id: int,
    update_msg: NotificationUpdate,
    db: Session = Depends(dependencies.get_db),
):
    db_notification = notifications.get(db, id=notification_id)
    db_notification = notifications.update(
        db=db, db_obj=db_notification, obj_in=update_msg
    )
    return db_notification


@router.delete("/{notification_id}", status_code=202, response_model=Notification)
def delete_notification(
    notification_id: int, db: Session = Depends(dependencies.get_db)
):
    deleted_notification = notifications.remove(db=db, id=notification_id)
    return deleted_notification


@router.get("/{notification_id}", response_model=Notification)
def get_notification(notification_id: int, db: Session = Depends(dependencies.get_db)):
    db_notification = notifications.get(db=db, id=notification_id)
    return db_notification


@router.get("", response_model=List[Notification])
def get_notifications(
    limit: int = 100, skip: int = 0, db: Session = Depends(dependencies.get_db)
):
    db_notifications = notifications.get_multi(db=db, skip=skip, limit=limit)
    return db_notifications
